;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;                                                                 ;;
;; Copyright (C) KolibriOS team 2004-2018. All rights reserved.    ;;
;; Distributed under terms of the GNU General Public License       ;;
;;                                                                 ;;
;;   Written by hidnplayr@kolibrios.org                            ;;
;;                                                                 ;;
;;         GNU GENERAL PUBLIC LICENSE                              ;;
;;          Version 2, June 1991                                   ;;
;;                                                                 ;;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;


user_parser:

        mov     ebp, [window_active]                    ; print to the current window
        mov     [window_print], ebp

        mov     ecx, [edit1.size]
        test    ecx, ecx
        jz      .ret                                    ; ignore empty commands

        mov     esi, input_text
        mov     edi, user_command
        call    recode                                  ; Convert to UTF-8

        cmp     byte[user_command], '/'                 ; is it a server command ?
        je      .dont_change_my_msg

        mov     esi, user_command
        call    change_my_msg                           ; expand/shift user message, if selected

    .dont_change_my_msg:

        mov     dword[edi], 0x00000a0d                  ; terminate the line
        sub     edi, user_command
        mov     [user_command.size], edi

        cmp     byte[user_command], '/'                 ; is it a server command ?
        je      .command

        cmp     [status], STATUS_LOGGED_IN
        jne     .not_loggedin

        cmp     [ebp + window.type], WINDOWTYPE_CHANNEL
        je      .send_privmsg
        cmp     [ebp + window.type], WINDOWTYPE_CHAT
        jne     .not_channel

  .send_privmsg:
        if TIMESTAMP
        call    print_timestamp
        end if

        mov     al, '<'
        call    print_char

        mov     esi, user_nick
        call    print_asciiz

        mov     al, '>'
        call    print_char
        mov     al, ' '
        call    print_char

        mov     eax, [user_command.size]
        mov     byte[user_command + eax],0
        mov     esi, user_command
        call    print_asciiz

        mov     al, 10
        call    print_char

; encrpt user message, if selected

        cmp byte[crpt_state], 0
        je .dont_crpt_my_msg

        push esi
        mov esi, user_command
        call crpt_my_msg
        pop esi

  .dont_crpt_my_msg:

; and now send it to the server
        mov     dword[packetbuf], 'PRIV'
        mov     dword[packetbuf+4], 'MSG '

        lea     esi, [ebp + window.name]
        mov     edi, packetbuf+8
        mov     ecx, MAX_WINDOWNAME_LEN
  .loop:
        lodsb
        test    al, al
        jz      .done
        stosb
        dec     ecx
        jnz     .loop
  .done:

        mov     ax, ' :'
        stosw

        mov     esi, user_command
        mov     ecx, [user_command.size]
;        inc     ecx
        rep     movsb

; end the command with a CRLF
        mov     ax, 0x0a0d
        stosw

        lea     esi, [edi - packetbuf]
        mcall   send, [socketnum], packetbuf, , 0

  .ret:
        ret

; Text begins with a '/' lets try to find the command in the lookup table.
  .command:
        mov     eax, dword[user_command+1]      ; skip '/'
        or      eax, 0x20202020                 ; convert to lowercase

        mov     edi, user_commands
        mov     ecx, user_commands.number
        cmp     [status], STATUS_CONNECTED
        jae     .cmd_loop
        mov     ecx, user_commands.number2
  .cmd_loop:
        scasd
        je      .got_cmd
        add     edi, 4
        dec     ecx
        jnz     .cmd_loop

        cmp     [status], STATUS_CONNECTED
        jb      .not_connected

; Commands shorter then 3 chars are placed here
        and     eax, 0x00ffffff
        cmp     eax, 'me '
        je      cmd_usr_me

; If none of the listed commands, send text straight to server
        jmp     cmd_usr_send

  .got_cmd:
        jmp     dword[edi]

  .not_loggedin:
        mov     esi, str_notloggedin
        call    print_asciiz
        ret

  .not_connected:
        mov     esi, str_notconnected
        call    print_asciiz
        ret

  .not_channel:
        mov     esi, str_notchannel
        call    print_asciiz
        ret


; user commands lookup table
user_commands:
        dd      'nick', cmd_usr_nick
        dd      'real', cmd_usr_real
        dd      'serv', cmd_usr_server
        dd      'help', cmd_usr_help
        dd      'code', cmd_usr_code
        dd      'crpt', cmd_usr_crpt
        dd      'expd', cmd_usr_expd
        dd      'shft', cmd_usr_shft

        .number2 = ($ - user_commands) / 8

; All following commands require a connection to the server.
        dd      'quer', cmd_usr_quer
        dd      'quit', cmd_usr_quit
        dd      'part', cmd_usr_part
        dd      'ctcp', cmd_usr_ctcp
        dd      'msg ', cmd_usr_msg

        .number = ($ - user_commands) / 8



cmd_usr_msg:

        lea     esi, [user_command+5]

        mov     dword[packetbuf], 'PRIV'
        mov     dword[packetbuf+4], 'MSG '
        lea     edi, [packetbuf+8]

        ; in addition, need to find (nickname_length + 1) and put it to ECX register

        push    ecx ebx
        xor     ebx, ebx
        xor     ecx, ecx

  @@:
        lodsb
        test    al, al
        jz      .fail
        cmp     al, 10
        je      .fail
        cmp     al, 13
        je      .fail
        inc     ecx
        stosb
        cmp     al, ' '
        jne     @r

        mov     al, ':'
        stosb

; expand/shift user message, if selected

        mov     ebx, [user_command.size]      ; total_length
        sub     ebx, ecx                      ; total_length - (nickname_length + 1)
        sub     ebx, 5                        ; ^ - ("/msg " length) = real_message_length

        push    edi

        mov     edi, esi
        add     edi, ebx

        call    change_my_msg                 ; expand/shift user message, if selected

        mov     dword[edi], 0x00000a0d

        pop     edi ebx ecx

        push    edi esi esi

; print to local window
        if TIMESTAMP
        call    print_timestamp
        end if

        mov     esi, msg_header
        call    print_asciiz

        mov     esi, packetbuf+8
        mov     bl, ' '
        call    print_string

        mov     al, '*'
        call    print_char

        mov     al, ' '
        call    print_char

        pop     esi
        call    print_asciiz
        pop     esi

; encrpt user message, if selected

        cmp byte[crpt_state], 0
        je .dont_crpt_my_msg

        ; pointer is already at esi
        call crpt_my_msg

  .dont_crpt_my_msg:

; copy the message itself
  @@:
        lodsb
        test    al, al
        jz      @f
        cmp     al, 10
        je      @f
        cmp     al, 13
        je      @f
        stosb
        jmp     @r
  @@:
; end the command with a CRLF
        mov     ax, 0x0a0d
        stosw
        mov     byte[edi], 0

; now send it to the server

        pop     esi

        lea     esi, [edi - packetbuf]
        mcall   send, [socketnum], packetbuf, , 0

  .fail:
        ret



cmd_usr_quit:

        mov     esi, quit_msg
        cmp     byte[user_command+5], ' '
        jne     quit_server
        lea     esi, [user_command+6]

; esi = quit message
quit_server:

; User wants to close a channel, send PART command to server
        mov     dword[packetbuf], 'QUIT'
        mov     word[packetbuf+4], ' :'
        lea     edi, [packetbuf+6]
; Append our quit msg
  @@:
        lodsb
        cmp     al, 13
        je      @f
        test    al, al
        jz      @f
        stosb
        jmp     @r
  @@:
; end the command with a CRLF
        mov     ax, 0x0a0d
        stosw

        lea     esi, [edi - packetbuf]                  ; calculate length
        mcall   send, [socketnum], packetbuf, , 0       ; and finally send to server

        mov     ebp, windows
  .window_loop:
        cmp     [ebp + window.type], WINDOWTYPE_NONE
        je      .next_window
        mov     [window_print], ebp
        if TIMESTAMP
        call    print_timestamp
        end if
        mov     esi, str_disconnected
        call    print_asciiz
        cmp     [ebp + window.type], WINDOWTYPE_CHANNEL
        jne     .next_window
        call    user_remove_all
  .next_window:
        add     ebp, sizeof.window
        cmp     ebp, windows + (MAX_WINDOWS * sizeof.window)
        jb      .window_loop

        mov     [status], STATUS_DISCONNECTED
        mcall   close, [socketnum]

        ret




cmd_usr_nick:

        cmp     [user_command.size], 5
        je      .justprint
        cmp     byte[user_command+5], ' '
        jne     .fail
        cmp     [socketnum], 0
        je      .dontsend

; Request nickname change to server
        mov     dword[user_command+1], 'NICK'
        mov     esi, [user_command.size]
        mov     word[user_command + esi], 0x0a0d
        inc     esi
        mcall   send, [socketnum], user_command+1, , 0

  .fail:
        ret

; We arent logged in yet, directly change user_nick field and print notification to user.
  .dontsend:
        mov     ecx, MAX_NICK_LEN
        mov     esi, user_command+6
        mov     edi, user_nick
  @@:
        lodsb
        cmp     al, 13
        je      @f
        stosb
        dec     ecx
        jnz     @r
  @@:
        xor     al, al
        stosb

  .justprint:
        mov     esi, str_nickchange
        call    print_asciiz

        mov     esi, user_nick
        call    print_asciiz

        mov     al, 10
        call    print_char

        ret



cmd_usr_real:

        cmp     byte[user_command+5], ' '
        jne     cmd_usr_send

        mov     ecx, MAX_REAL_LEN
        mov     esi, user_command+6
        mov     edi, user_real_name
  .loop:
        lodsb
        cmp     al, 13
        je      .done
        stosb
        dec     ecx
        jnz     .loop
  .done:
        xor     al, al
        stosb

        mov     esi, str_realchange
        call    print_asciiz

        mov     esi, user_real_name
        call    print_asciiz

        mov     al, 10
        call    print_char

        ret



cmd_usr_server:

        mov     eax, dword[user_command+5]      ; check for 'er ', we only checked 'serv'
        or      eax, 0x00002020
        and     eax, 0x00ffffff
        cmp     eax, 'er '
        jne     cmd_usr_send

; Server window is always first window in the list, switch to it.
        mov     [window_print], windows
        mov     [window_active], windows

        mov     ecx, [user_command.size]        ; ok now set the address
        sub     ecx, 8

        mov     esi, user_command+8
  .now:
        push    esi
        mov     edi, irc_server_name
  .loop:                                        ; copy until zero byte, or ecx reaches zero.
        lodsb
        stosb
        test    al, al
        jz      .done
        dec     ecx
        jnz     .loop
        xor     al, al
        stosb
  .done:
        pop     esi

; set it also in window name
        mov     ebx, [window_print]
        call    window_set_name

; now connect
        call    socket_connect

        ret


cmd_usr_quer:

        mov     esi, user_command+7
        call    window_open
;        test    ebx, ebx
;        jz      .fail

        ret



cmd_usr_help:

        mov     esi, str_help
        call    print_asciiz

        ret



cmd_usr_code:

        ; TODO

        ret


; User wants to enable or disable the encrption

cmd_usr_crpt:

    push esi ecx eax

    mov esi, str_crptstate
    call print_asciiz

    cmp byte [crpt_state], 0
    je .crpt_enable         ; if disabled, lets see if we could enable
                            ; else, if enabled - we are going to disable

    call clear_crpt_stuff   ; erase RAM areas which could contain the crpt stuff

    mov byte [crpt_state], 0
    mov byte [expd_state], 0
    mov byte [shft_state], 0

    mov esi, str_disabled
    call print_asciiz

    mov [edit1.max], USERCMD_MAX_SIZE      ; max size of input textbox

    jmp .usr_crpt_done

  .crpt_enable:

    ; check if the user-specified encrption key is meeting the requirements

    call crpt_check_key
    cmp ecx, 0
    jne .usr_crpt_done      ; if not 0 there is a problem with key, dont enable

    ; enable the encrption and print a couple of messages

    mov byte [crpt_state], 1
    mov byte [expd_state], 1
    mov byte [shft_state], 1

    mov esi, str_enabled
    call print_asciiz

    mov esi, str_crptnote
    call print_asciiz

    ; copy the encrption key from the user_command (skipping "/crpt ") to its storage

    mov     ecx, CRPT_KEY_LEN
    mov     esi, user_command+6
    mov     edi, crpt_key
  @@:
        lodsb
        cmp     al, 13
        je      @f
        stosb
        dec     ecx
        jnz     @r
  @@:
        xor     al, al
        stosb

    mov edi, user_command
    mov ebx, MAX_COMMAND_LEN_BYTES
    call clearmem

    mov [edit1.max], CRPT_KEY_LEN      ; max size of input textbox

  .usr_crpt_done:

    mov al, 10
    call print_char

    pop eax ecx esi

    ret

; crpt_check_key:
; check if the user-specified encrption key is meeting the requirements
;
; IN:   ECX  = ptr to user_command with encrption key
; OUT:  ECX  = 0 if valid,
;            = 1 if any symbols do not belong to 32-126 code range,
;           >= 6 (keylen + 6) if a key is too short or too long

crpt_check_key:

    mov ecx, 6     ; skip "/crpt "

  .crpt_find_keylen_cycle:

    mov al, byte[user_command + ecx]

    cmp al, 0x0d      ; if its the end of input, quit a cycle
    je .crpt_findkeylen_done
    cmp al, 0x0a      ; if its the end of input, quit a cycle
    je .crpt_findkeylen_done

    cmp al, 0x7e      ; check if a symbol is above 32-126 code range
    ja .crpt_not7bitasciirange
    cmp al, 0x20      ; check if a symbol is below 32-126 code range
    jb .crpt_not7bitasciirange

    inc ecx
    jmp .crpt_find_keylen_cycle

  .crpt_not7bitasciirange:

    mov esi, str_not7bitasciirange
    call print_asciiz

    mov ecx, 1
    ret          ; error code = 1

  .crpt_findkeylen_done:

    cmp ecx, CRPT_KEY_LEN+6      ; +6 because of "/crpt "
    je .check_key_done

    mov esi, str_keylenerror
    call print_asciiz

    ret          ; error code = keylen + 6

  .check_key_done:

    mov ecx, 0      ; no error

    ret

; crpt_my_msg:
; encrpt my message, pointer to which is at ESI
;
; IN:   ESI = ptr to source message
; OUT:  ESI = ptr to encrpted message

crpt_my_msg:

    push esi ebx eax

    xor eax, eax

  .cycle_crpt_my_msg:

    cmp byte[esi], 0x01
    je .exit_cycle_crpt_my_msg
    cmp byte[esi], 0x0a
    je .exit_cycle_crpt_my_msg
    cmp byte[esi], 0x0d
    je .exit_cycle_crpt_my_msg

    cmp byte[esi], 0x7e ; check if a symbol is above 32-126 code range
    ja .crpt_go_to_next_char
    cmp byte[esi], 0x20 ; check if a symbol is below 32-126 code range
    jb .crpt_go_to_next_char

    mov bl, byte[crpt_key + eax]
    sub bl, 31
    add byte[esi], bl
    cmp byte[esi], 127
    jb .no_overflow

    sub byte[esi], 126
    add byte[esi], 31

  .no_overflow:

    inc eax

  .crpt_go_to_next_char:

    inc esi

    cmp eax, CRPT_KEY_LEN
    jne .cycle_crpt_my_msg

  .exit_cycle_crpt_my_msg:

    pop eax ebx esi

    ret

; change_my_msg:
; expand/shift user message, if selected
;
; IN:       ESI = ptr to the beginning of source message
;           EDI = ptr to the end of source message

change_my_msg:

        push    edx ecx

        xor     edx, edx
        add     dl, byte[expd_state]
        add     dl, byte[shft_state]

        cmp     edx, 0
        je      .dont_change_my_msg

        ; put the length of a message into EDX register

        mov     edx, edi
        sub     edx, esi

        ; EDX = length of a message (input argument for shift_my_msg if called)

        call    fill_random_buffer
        mov     ecx, RANDOM_BUFFER_SIZE_TO_CONVERT
        call    convert_random_buffer_to_ascii

        cmp     byte[expd_state], 0
        je      .dont_expand_my_msg

        call    expand_my_msg
        mov     edx, CRPT_KEY_LEN

    .dont_expand_my_msg:

        cmp     byte[shft_state], 0
        je      .dont_shift_my_msg

        call    shift_my_msg

    .dont_shift_my_msg:

    .dont_change_my_msg:

        pop ecx edx

        ret


; User wants to enable or disable the expansion
cmd_usr_expd:
    push esi
    mov esi, str_expdstate
    call print_asciiz
    cmp byte[expd_state], 0
    je .expd_enable
    jmp .expd_disable
  .expd_enable:
    mov byte[expd_state], 1
    mov esi, str_enabled
    call print_asciiz
    jmp .exit_expd
  .expd_disable:
    mov byte[expd_state], 0
    mov esi, str_disabled
    call print_asciiz
  .exit_expd:
    pop esi
    ret

; expand_my_msg:
; expand my message, pointer to which is at ESI, with random bytes
;
; IN:       ESI = ptr to the beginning of source message
;           EDI = ptr to the end of source message
;
; NOTE:     this function depends on random, before calling this function it is required to:
;
;               call    fill_random_buffer
;               call    convert_random_buffer_to_ascii

expand_my_msg:

    push esi edx ecx ebx eax

    mov ebx, esi
    add ebx, CRPT_KEY_LEN

    cmp ebx, edi
    jna .cant_expand

    sub ebx, edi

    xor eax, eax
    xor ecx, ecx
    mov edx, random_buffer

  .cycle_expand_my_msg_with_random_buf:

    mov al, byte[edx]
    stosb

    inc edx
    inc ecx

    cmp ecx, ebx
    jne .cycle_expand_my_msg_with_random_buf

  .cant_expand:

    pop eax ebx ecx edx esi

    ret

; User wants to enable or disable the shifting
cmd_usr_shft:
    push esi
    mov esi, str_shftstate
    call print_asciiz
    cmp byte[shft_state], 0
    je .shft_enable
    jmp .shft_disable
  .shft_enable:
    mov byte[shft_state], 1
    mov esi, str_enabled
    call print_asciiz
    jmp .exit_shft
  .shft_disable:
    mov byte[shft_state], 0
    mov esi, str_disabled
    call print_asciiz
  .exit_shft:
    pop esi
    ret

; shift_my_msg:
; shift my message, pointer to which is at ESI, "clockwise" (right shift) to a random offset
;
; IN/OUT:   ESI = ptr to the beginning of source message
;           EDX = length of source message
;
; NOTE:     this function depends on random, before calling this function it is required to:
;
;               call    fill_random_buffer
;               call    convert_random_buffer_to_ascii

shift_my_msg:

    push edi esi ecx ebx eax

    ; check if a length of source message is between the supported limits

    cmp edx, 2
    jb .exit_shift_my_msg
    cmp edx, CRPT_KEY_LEN
    ja .exit_shift_my_msg

    ; get a value of random offset, and move it to EBX register

    xor eax, eax

    mov ebx, random_buffer
    add ebx, edx
    inc ebx

    mov al, byte[ebx]

    ; random offset should not be larger than a message length,
    ; so if its larger: RANDOM OFFSET = RANDOM_VALUE mod MESSAGE_LENGTH

    xor ebx, ebx

    cmp eax, edx
    jb .skip_division

    div dl
    mov al, ah

  .skip_division:

    mov bl, al

    ; copy the right part of message (starting with EBX offset) to the temporary buffer

    xor eax, eax

    mov ecx, edx
    sub ecx, ebx

    mov edi, msg_temp_buf
    add edi, ebx

    call shift_my_msg_cycle

    ; copy the left part of message (before the ebx offset) to the temporary buffer,
    ; placing it after the right part which we have copied previously

    mov edi, msg_temp_buf
    mov ecx, edx

    call shift_my_msg_cycle

    ; copy the temporary buffer, which contains a "clockwise" right shifted message,
    ; replacing the original message

    mov edi, esi
    sub edi, edx
    mov esi, msg_temp_buf
    mov ebx, edx

    call copymem

    ; check if the ending symbol of user_command is not a header for 2-byte char

    dec edi

    cmp byte[edi], 0xd7
    ja .exit_shift_my_msg
    cmp byte[edi], 0xd0
    jb .exit_shift_my_msg

    ; this 2-byte char will be splitted and corrupted by this shifting operation,
    ; we could only replace it with a couple of random chars to avoid the problems

    mov eax, random_buffer
    add eax, edx

    ; replace the header with random_buffer[message_length-1] unused random character

    dec eax
    mov bl, byte[eax]
    mov byte[edi], bl

    ; replace the letter with random_buffer[message_length  ] unused random character

    inc eax
    mov bl, byte[eax]
    sub edi, edx
    inc edi
    mov byte[edi], bl

  .exit_shift_my_msg:

    pop eax ebx ecx esi edi

    ret

; shift_my_msg_cycle:
; helper cycle for a shift_my_msg function, copies the left or right part of the message
;
; IN/OUT:   ESI = ptr to the source
;           EAX = current offset from the beginning of source
;           ECX = offset from the beginning of source, where to stop the copying
;           EDI = ptr to the destination

shift_my_msg_cycle:

  .cycle_shift_my_msg:

    cmp eax, ecx
    je .exit_cycle_shift_my_msg

    mov bl, byte[esi]
    mov byte[edi], bl

    inc eax
    inc esi
    inc edi

    jmp .cycle_shift_my_msg

  .exit_cycle_shift_my_msg:

    ret

; copymem:
; copy a memory area
;
; IN:   ESI - ptr to source memory area
;       EDI - ptr to destination memory area
;       EBX - how many bytes to copy

copymem:

    push eax ecx

    xor ecx, ecx

    .cycle_copymem:

        lodsb
        stosb
        inc ecx
        cmp ecx, ebx
        jb .cycle_copymem

    pop ecx eax

    ret

; clearmem:
; clear a memory area
;
; IN:   EDI - pointer to a memory area
;       EBX - how many bytes to clear

clearmem:

    push eax ecx

    xor al, al
    xor ecx, ecx

    .cycle_clearmem:

        stosb
        inc ecx
        cmp ecx, ebx
        jb .cycle_clearmem

    pop ecx eax

    ret

; clear_crpt_stuff:
; erase RAM areas which could contain the crpt stuff

clear_crpt_stuff:

    push edi ebx

    mov edi, crpt_key
    mov ebx, CRPT_KEY_LEN
    call clearmem

    mov edi, user_command
    mov ebx, MAX_COMMAND_LEN_BYTES
    call clearmem
    mov edi, msg_temp_buf
    call clearmem

    mov edi, input_text
    mov ebx, MAX_COMMAND_LEN
    call clearmem

    mov edi, packetbuf
    mov ebx, PACKETBUF_SIZE
    call clearmem

    mov edi, path
    mov ebx, PATH_SIZE
    call clearmem

    mov edi, param
    mov ebx, PARAM_SIZE
    call clearmem

    mov edi, servercommand
    mov ebx, SERVERCOMMAND_SIZE
    call clearmem

    mov edi, random_buffer
    mov ebx, RANDOM_BUFFER_SIZE
    call clearmem

    pop ebx edi

    ret

; User typed a part command
cmd_usr_part:

        cmp     byte[user_command+5], 13        ; parameters given?
        jne     cmd_usr_send                    ; yes, send command straight to server

; close active window
cmd_usr_close_window:

        mov     esi, [window_active]
        mov     [window_print], esi
        cmp     [esi + window.type], WINDOWTYPE_SERVER
        je      .not_channel

        lea     esi, [esi + window.name]
        call    cmd_usr_part_channel
        call    window_close
        ret

  .not_channel:
        cmp     [esi + window.type], WINDOWTYPE_CHAT
        jne     .not_chat

        call    window_close
  .not_chat:

        ret



; Send part command to server
; esi must point to channel name (ASCIIZ)
cmd_usr_part_channel:

; User wants to close a channel, send PART command to server
        mov     dword[packetbuf], 'PART'
        mov     byte[packetbuf+4], ' '
        lea     edi, [packetbuf+5]
  @@:
        lodsb
        test    al, al
        jz      @f
        cmp     al, 13
        je      @f
        cmp     al, 10
        je      @f
        stosb
        jmp     @r
  @@:
; end the command with a CRLF
        mov     ax, 0x0a0d
        stosw

        lea     esi, [edi - packetbuf]                  ; calculate length
        mcall   send, [socketnum], packetbuf, , 0       ; and finally send to server

        ret



cmd_usr_ctcp:

        cmp     byte[user_command+5], ' '
        jne     cmd_usr_send

        mov     esi, user_command+6
; prepare a 'PRIVMSG '
        mov     dword[packetbuf], 'PRIV'
        mov     dword[packetbuf+4], 'MSG '
        lea     edi, [packetbuf+8]

        ; in addition, need to find ("nickname/channel length" + 1) and put it to ECX register

        push    ecx ebx
        xor     ebx, ebx
        xor     ecx, ecx

; append the destination (nickname/channel)
  @@:
        lodsb
        test    al, al
        jz      .fail
        cmp     al, ' '
        je      @f
        inc     ecx
        stosb
        jmp     @r
  @@:

        mov     ax, ' :'
        stosw
        mov     al, 0x01
        stosb

; expand/shift user message, if selected

        inc     ecx                           ; +1 for a space character after the nickname/channel

        mov     ebx, [user_command.size]      ; total_length
        sub     ebx, ecx                      ; total_length - ("nickname/channel length" + 1)
        sub     ebx, 6                        ; ^ - ("/ctcp " length) = real_message_length

        push    edi

        mov     edi, esi
        add     edi, ebx

        call    change_my_msg                 ; expand/shift user message, if selected

        mov     dword[edi], 0x00000a0d

        pop     edi ebx ecx

        push    edi esi esi

; print to local window
        if TIMESTAMP
        call    print_timestamp
        end if

        mov     esi, ctcp_header
        call    print_asciiz

        mov     esi, packetbuf+8
        mov     bl, ' '
        call    print_string

        mov     al, ']'
        call    print_char

        mov     al, ' '
        call    print_char

        pop     esi
        call    print_asciiz
        pop     esi

; encrpt user message, if selected

        cmp byte[crpt_state], 0
        je .dont_crpt_my_msg

        ; pointer is already at esi
        call crpt_my_msg

  .dont_crpt_my_msg:

; copy the message itself
  @@:
        lodsb
        test    al, al
        jz      @f
        cmp     al, 13
        je      @f
        cmp     al, 10
        je      @f
        stosb
        jmp     @r
  @@:

; end of message
        mov     al, 0x01
        stosb
        mov     ax, 0x0a0d
        stosw
        mov     byte[edi], 0

; now send it to the server

        pop     esi

        lea     esi, [edi - packetbuf]                  ; calculate length
        mcall   send, [socketnum], packetbuf, , 0       ; and finally send to server

  .fail:

        ret



cmd_usr_me:

; prepare a 'PRIVMSG '
        mov     dword[packetbuf], 'PRIV'
        mov     dword[packetbuf+4], 'MSG '
        lea     edi, [packetbuf+8]

; append the destination (nickname/channel)
        mov     esi, [window_active]
        lea     esi, [esi + window.name]
  @@:
        lodsb
        test    al, al
        je      @f
        stosb
        jmp     @r
  @@:

; Make the CTCP action header
        mov     eax, ' :' + 0x01 shl 16 + 'A' shl 24
        stosd
        mov     eax, 'CTIO'
        stosd
        mov     al, 'N'
        stosb

; expand/shift user message, if selected

        mov     esi, user_command+4           ; skip "/me "

        push    ebx
        xor     ebx, ebx

        mov     ebx, [user_command.size]      ; total_length
        sub     ebx, 4                        ; ^ - ("/me " length) = real_message_length

        push    edi

        mov     edi, esi
        add     edi, ebx

        call    change_my_msg                 ; expand/shift user message, if selected

        mov     dword[edi], 0x00000a0d

        pop     edi ebx

; print to local window
        if TIMESTAMP
        call    print_timestamp
        end if

        mov     esi, action_header
        call    print_asciiz

        mov     esi, user_nick
        call    print_asciiz

        mov     esi, user_command+3
        mov     bl, 13
        call    print_string

        mov     al, 10
        call    print_char

; encrpt user message, if selected

        cmp byte[crpt_state], 0
        je .dont_crpt_my_msg

        mov esi, user_command+4
        call crpt_my_msg

  .dont_crpt_my_msg:

; copy the message itself (including first space)
        mov     esi, user_command+3
  @@:
        lodsb
        cmp     al, 13
        je      @f
        stosb
        jmp     @r
  @@:

; end of CTCP message
        mov     al, 0x01
        stosb
        mov     ax, 0x0a0d
        stosw
        mov     byte[edi], 0

; now send it to the server
        lea     esi, [edi - packetbuf]                  ; calculate length
        mcall   send, [socketnum], packetbuf, , 0       ; and finally send to server

        ret



; The user typed some undefined command, just send it to the server
cmd_usr_send:

        mov     esi, [user_command.size]
        mov     eax, [user_command.size]
        add     eax, user_command+1
        mov     word[eax], 0x0d0a
        inc     esi
        mcall   send, [socketnum], user_command+1, , 0

        ret





